(define-macro (when test . forms)
  `(',if ,test (',begin ,@forms)))

(define-macro (unless test . forms)
  `(',if (',not ,test) (',begin ,@forms)))

(define (values&forms-1 form)
  `(',cons ,form ',form))
(define-macro (values&forms . forms)
  "\
Return a list of pairs, one per argument.  Each pair contains the value of
the argument as the car and the unevaluated expression as the cdr.
(values&forms (* 3 5) (string-append \"foo\" \"bar\"))
=> ((15 . (* 3 5)) (\"foobar\" . (string-append \"foo\" \"bar\")))"
  `(',list ,@(map values&forms-1 forms)))

(define (qw-1 form)
  (cond
   ((string? form) `'(,form))
   ((symbol? form) `'(,(symbol->string form)))
   (else
    `(',let ((x ,form))
       (',if (',list? x)
         x
         (',list x))))))
(define-macro (qw . forms)
  "\
Return a list of strings, analogous to Perl's qw().  Each argument can be a
string, which is included as-is; a symbol, whose name is used; or an
expression, which is evaluated.  If the expression yields a list value, the
elements of that list are interpolated directly into the returned list.
(qw \"foo\" 'bar (list \"baz\" \"quux\"))
=> (\"foo\" \"bar\" \"baz\" \"quux\")"
  `(',append . ,(map qw-1 forms)))

(define-macro (scope . forms)
  (define bindings (make-q))
  (define body (make-q))
  (for-each
   (lambda (form)
     (if (and (pair? form) (eq? (car form) 'defn))
       (begin
         (while (not (q-empty? body))
           (enq! bindings (list (gensym) (q-pop! body))))
         (if (assq-ref (car bindings) (cadr form))
           (throw 'misc-error "scope" "duplicate bindings" '() #f))
         (enq! bindings (cdr form)))
       (enq! body form)))
   forms)
  (if (null? (car bindings))
    `(',begin . ,(car body))
    `(',let* ,(car bindings) . ,(car body))))

(define-macro (defproc prototype . forms)
  "\
Define a procedure, like (define (...) ...), but also return the procedure."
  `(',let ()
     (define ,prototype . ,forms)
     ,(car prototype)))

(define-macro (defproc* prototype . forms)
  "\
Define a procedure, like (define* (...) ...), but also return the procedure."
  `(',let ()
     (define* ,prototype . ,forms)
     ,(car prototype)))
